/*
 * ctmonitor.c	"ct monitor"
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Authors:	Patrick McHardy, <kaber@trash.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include "ct.h"
#include "utils.h"

static void usage(void)
{
	fprintf(stdout, "Usage: ct monitor [ sockbufsize SIZE | mcgroup CTNLGROUP[,..] ]\n");
	fprintf(stdout, "       ct monitor network [ port PORT | mcgroup IP ]\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "CTNLGROUP := { tcp | udp | icmp | other }\n");
	fprintf(stdout, "\n");
}

static void explain(char *arg)
{
	fprintf(stderr, "Illegal \"%s\"\n", arg);
}

static int accept_msg(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	print_nlmsg((FILE *)arg, n);
	return 0;
}

int ctmonitor(int argc, char **argv)
{
	struct nfnl_handle nlh;
	unsigned int mcgroup = 0;
	unsigned short port = 0;
	int network = 0, sock, sock_buf_size = NETLINK_SOCKBUFSZ;

	while (argc > 0) {
		if (matches(*argv, "network")) {
			network = 1;
		} else if (matches(*argv, "mcgroup")) {
			NEXT_ARG();
			if (network) {
				if (get_addr(&mcgroup, *argv)) {
					explain("mcgroup");
					return -1;
				}
			} else {
				if (get_nlmcgroups(&mcgroup, *argv)) {
					explain("mcgroup");
					return -1;
				}
			}
		} else if (matches(*argv, "port")) {
			if (!network) {
				fprintf(stderr, "Port can only be specified "
						"in network mode.\n");
				return -1;
			}
			NEXT_ARG();
			if (get_u16(&port, *argv, 0)) {
				explain("port");
				return -1;
			}
		} else if (matches(*argv, "sockbufsize")) {
			NEXT_ARG();
			if (get_size(&sock_buf_size, *argv)) {
				explain("sockbufsize");
				return -1;
			}
		} else if (matches(*argv, "help")) {
			usage();
			return -1;
		} else {
			fprintf(stderr, "Argument \"%s\" is unknown, try "
					"\"ct monitor help\".\n", *argv);
			return -1;
		}
		argv++, argc--;
	}

	
	if (network) {
		char buf[4000];
		struct nlmsghdr *n = (struct nlmsghdr *)buf;
		struct sockaddr_in sin_l;
		int len;
		
		if (!port)
	 		port = CTSYNCD_DEFPORT;
		sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (sock < 0) {
			perror("socket");
			return -1;
		}
		sin_l.sin_family      = AF_INET;
		sin_l.sin_port        = htons(port);
		sin_l.sin_addr.s_addr = INADDR_ANY;
		if (bind(sock, (struct sockaddr *)&sin_l, sizeof(sin_l)) < 0) {
			perror("bind");
			return -1;
		}
		
		if (mcgroup) {
			struct ip_mreq mreq = {
				.imr_multiaddr.s_addr = mcgroup,
				.imr_interface.s_addr = INADDR_ANY
			};
			if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			               &mreq, sizeof(mreq)) < 0) {
				perror("setsockopt");
				return -1;
			}
		}

		while ((len = recvfrom(sock, &buf, sizeof(buf), 0, NULL, 0)))
			print_nlmsg(stdout, n);
		/* ONLY REACHED ON ERROR */
		return -1;
	}

	if (!mcgroup)
		mcgroup = ~0;
	
	if (nfnl_open(&nlh, NFNL_SUBSYS_CTNETLINK, mcgroup))
		return -1;
	
	if (setsockopt(nlh.fd, SOL_SOCKET, SO_RCVBUF,
	               (const char *)&sock_buf_size, sizeof(int)) < 0) {
		perror("setsockopt");
		return -1;
	}

	if (nfnl_listen(&nlh, accept_msg, stdout) < 0)
		return -1;

	exit(0);
}
