/*
 * ctsycnd.c    "ct syncd"
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Authors:     Patrick McHardy, <kaber@trash.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include "ct.h"
#include "utils.h"

struct mc_bundle {
	int fd;
	struct sockaddr *sa_to;
	int sa_len;
};

static void usage(void)
{
	fprintf(stdout, "Usage: ct syncd [ listen ] [ ip IP | port PORT ]\n");
}

static void explain(char *arg)
{
	fprintf(stderr, "Illegal \"%s\"\n", arg);
}

static int accept_msg(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	struct mc_bundle *b = (struct mc_bundle *)arg;

	print_nlmsg(stdout, n);
	if (sendto(b->fd, n, n->nlmsg_len, 0, b->sa_to, b->sa_len) < 0)
		perror("send");

	return 0;
}

static int accept_msg_junk(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	fprintf(stdout, "JUNK: SEQ=%u\n", n->nlmsg_seq);
	if (n->nlmsg_flags == NLMSG_ERROR)
		fprintf(stdout, "ERROR\n");
	print_nlmsg(stdout, n);
	return 0;
}

int ctsyncd(int argc, char **argv)
{
	struct nfnl_handle nlh;
	struct mc_bundle mc;
	struct sockaddr_in sin;
	int sin_len = sizeof(struct sockaddr_in);
	int sock_buf_size = NETLINK_SOCKBUFSZ;
	int sock;
	int listen = 0;
	u_int16_t port = CTSYNCD_DEFPORT;
	u_int32_t addr = 0;

	while (argc > 0) {
		if (matches(*argv, "listen")) {
			listen = 1;
		} else if (matches(*argv, "ip")) {
			NEXT_ARG();
			if (get_addr(&addr, *argv)) {
				explain("ip");
				return -1;
			}
		} else if (matches(*argv, "port")) {
			NEXT_ARG();
			if (get_u16(&port, *argv, 0)) {
				explain("port");
				return -1;
			}
		} else if (matches(*argv, "help")) {
			usage();
			return -1;
		} else {
			fprintf(stdout, "error\n");
			return -1;
		}
		argv++, argc--;
	}

	if (nfnl_open(&nlh, NFNL_SUBSYS_CTNETLINK, 0)) {
		perror("nfnl_open");
		return -1;
	}

	if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		perror("socket");
		return -1;
	}

	sin.sin_family		= AF_INET;
	sin.sin_addr.s_addr	= addr;
	sin.sin_port		= htons(port);

	if (MULTICAST(addr)) {
		struct ip_mreq mreq = {
			.imr_multiaddr.s_addr = addr,
			.imr_interface.s_addr = INADDR_ANY
		};
		if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		               &mreq, sizeof(mreq)) < 0) {
			perror("setsockopt");
			return -1;
		}
	}

	if (listen) {
		char buf[4000];
		struct nlmsghdr *n = (struct nlmsghdr *)buf;
		int len;

		sin.sin_addr.s_addr = INADDR_ANY;
		if (bind(sock, (struct sockaddr *)&sin, sin_len) < 0) {
			perror("bind");
			exit(1);
		}

		while ((len = recvfrom(sock, &buf, sizeof(buf), 0, NULL, 0))) {
			n->nlmsg_flags |= NLM_F_REQUEST;
			if (nfnl_talk(&nlh, n, 0, 0, NULL, accept_msg_junk,
			              NULL) < 0)
				print_nlmsg(stdout, n);
		}

		exit(0);	/* NEVER REACHED */
	}

	if (sock_buf_size > get_sock_rmem_max())
		fprintf(stderr, "Warning: max socket rmem (%u) is less than "
		                "requested size (%u).\n",
				get_sock_rmem_max(), sock_buf_size);
	
	if (setsockopt(nlh.fd, SOL_SOCKET, SO_RCVBUF,
	               (const char *)&sock_buf_size, sizeof(int)) < 0) {
		perror("setsockopt");
		exit(1);
	}

	mc.fd     = sock;
	mc.sa_to  = (struct sockaddr *)&sin;
	mc.sa_len = sin_len;

	if (nfnl_listen(&nlh, accept_msg, &mc) < 0) {
		perror("nfnl_listen");
		exit(1);
	}

	exit(0);
}
