/*
 * print_nlmsg.c
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Authors:     Patrick McHardy, <kaber@trash.net>
 */

#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "ct.h"

static int print_invalid(FILE *, char *, void *);
static int print_tuple(FILE *, char *, void *);
static int print_status(FILE *, char *, void *);
static int print_protoinfo(FILE *, char *, void *);
static int print_helpinfo(FILE *, char *, void *);
static int print_natinfo(FILE *, char *, void *);
static int print_timeout(FILE *, char *, void *);
static int print_mark(FILE *, char *, void *);

struct ct_attribute {
	char *name;
	unsigned int size;
	int (*print)(FILE *, char *, void *);
};

static struct ct_attribute cta_table[CTA_MAX + 1] = {
	[CTA_UNSPEC] = {
		.name  = "CTA_UNSPEC",
		.size  = 0,
		.print = print_invalid
	},
	[CTA_ORIG] = {
		.name  = "CTA_ORIG",
		.size  = sizeof(struct ip_conntrack_tuple),
		.print = print_tuple
	},
	[CTA_RPLY] = {
		.name  = "CTA_RPLY",
		.size  = sizeof(struct ip_conntrack_tuple),
		.print = print_tuple
	},
	[CTA_STATUS] = {
		.name  = "CTA_STATUS",
		.size  = sizeof(unsigned long),
		.print = print_status
	},
	[CTA_PROTOINFO] = {
		.name  = "CTA_PROTOINFO",
		.size  = sizeof(struct cta_proto),
		.print = print_protoinfo
	},
	[CTA_HELPINFO] = {
		.name  = "CTA_HELPINFO",
		.size  = sizeof(struct cta_help),
		.print = print_helpinfo
	},
	[CTA_NATINFO] = {
		.name  = "CTA_NATINFO",
		.size  = sizeof(struct cta_nat),
		.print = print_natinfo
	},
	[CTA_TIMEOUT] = {
		.name  = "CTA_TIMEOUT",
		.size  = sizeof(unsigned long),
		.print = print_timeout
	},
	[CTA_MARK] = {
		.name  = "CTA_MARK",
		.size  = sizeof(unsigned long),
		.print = print_mark
	},
	[CTA_EXP_TIMEOUT] = {
		.name  = "CTA_EXP_TIMEOUT",
		.print = print_invalid
	},
	[CTA_EXP_TUPLE] = {
		.name  = "CTA_EXP_TUPLE",
		.print = print_invalid
	},
	[CTA_EXP_MASK] = {
		.name  = "CTA_EXP_MASK",
		.print = print_invalid
	},
	[CTA_EXP_SEQNO] = {
		.name  = "CTA_EXP_SEQNO",
		.print = print_invalid
	},
	[CTA_EXP_PROTO] = {
		.name = "CTA_EXP_PROTO",
		.print = print_invalid,
	},
	[CTA_EXP_HELP] = {
		.name = "CTA_EXP_HELP",
		.print = print_invalid,
	}
};

static int print_invalid(FILE *fp, char *type, void *data)
{
	fprintf(fp, "invalid type: %s\n", type);
	return 0;
}

static int print_tuple(FILE *fp, char *type, void *data)
{
	struct ip_conntrack_tuple *tuple = data;
	struct protoent *p;
	char proto[16];

	if ((p = getprotobynumber(tuple->dst.protonum)))
		strncpy(proto, p->p_name, sizeof(proto));
	else
		snprintf(proto, sizeof(proto), "%u", tuple->dst.protonum);

	fprintf(fp, "%s: %s %s:", type, proto,
	        inet_ntoa((struct in_addr){tuple->src.ip}));

	switch (tuple->dst.protonum) {
	case IPPROTO_ICMP:
		fprintf(fp, "ICMP (id %u)", ntohs(tuple->src.u.icmp.id));
		break;
	case IPPROTO_UDP:
	case IPPROTO_TCP:
		fprintf(fp, "%u", ntohs(tuple->src.u.tcp.port));
		break;
	default:
		fprintf(fp, "unknown");
		break;
	}

	fprintf(fp, "->%s:", inet_ntoa((struct in_addr){tuple->dst.ip}));

	switch (tuple->dst.protonum) {
	case IPPROTO_ICMP:
		fprintf(fp, "ICMP (type %u, code %u)",
		        tuple->dst.u.icmp.type, tuple->dst.u.icmp.code);
		break;
	case IPPROTO_UDP:
	case IPPROTO_TCP:
		fprintf(fp, "%u", ntohs(tuple->dst.u.tcp.port));
		break;
	default:
		fprintf(fp, "unknown");
		break;
	}

	fprintf(fp, "\n");

	return 0;
}

static int print_status(FILE *fp, char *type, void *data)
{
	unsigned long status = *(unsigned long *)data;
	char *delim = "";

	fprintf(fp, "%s: ", type);

	if (status & IPS_EXPECTED) {
		fprintf(fp, "%sexpected", delim);
		delim = ",";
	}
	if (status & IPS_CONFIRMED) {
		fprintf(fp, "%sconfirmed", delim);
		delim = ",";
	}
	if (status & IPS_SEEN_REPLY) {
		fprintf(fp, "%sseen reply", delim);
		delim = ",";
	}
	if (status & IPS_ASSURED) {
		fprintf(fp, "%sassured", delim);
		delim = ",";
	}
	fprintf(fp, "\n");

	return 0;
}

static inline int print_icmpinfo(FILE *fp, struct ip_ct_icmp *icmp)
{
	/* need cast from atomic_t */
	fprintf(fp, "count %u", *(unsigned int *)&icmp->count);
	return 0;
}

static inline int print_tcpinfo(FILE *fp, struct ip_ct_tcp *tcp)
{
	static char *tcp_states[TCP_CONNTRACK_MAX + 1] = {
		"NONE",
		"ESTABLISHED",
		"SYN_SENT",
		"SYN_RECV",
		"FIN_WAIT",
		"TIME_WAIT",
		"CLOSE",
		"CLOSE_WAIT",
		"LAST_ACK",
		"LISTEN",
		"INVALID"
	};

	if (tcp->state > TCP_CONNTRACK_MAX)
		tcp->state = TCP_CONNTRACK_MAX;

	fprintf(fp, "state %s ", tcp_states[tcp->state]);
	if (tcp->handshake_ack)
		fprintf(fp, "handshake ack %u ", ntohl(tcp->handshake_ack));
	
	return 0;
}

static int print_protoinfo(FILE *fp, char *type, void *data)
{
	struct cta_proto *cp = data;
	struct protoent *p = getprotobynumber(cp->num_proto);
	
	fprintf(fp, "%s: %s ", type, p->p_name);

	switch (cp->num_proto) {
	case IPPROTO_ICMP:
		print_icmpinfo(fp, &cp->proto.icmp);
		break;
	case IPPROTO_TCP:
		print_tcpinfo(fp, &cp->proto.tcp);
		break;
	}
	fprintf(fp, "\n");

	return 0;
}

static int print_helpinfo(FILE *fp, char *type, void *data)
{
	struct cta_help *ch = data;
	
	fprintf(fp, "%s: helper %s\n", type, ch->name);
	return 0;
}

static int print_natinfo(FILE *fp, char *type, void *data)
{
	static char *hooks[NF_IP_NUMHOOKS] = {
		"NF_IP_PRE_ROUTING",
		"NF_IP_LOCAL_IN",
		"NF_IP_FORWARD",
		"NF_IP_LOCAL_OUT",
		"NF_IP_POST_ROUTING"
	};

	struct cta_nat *nat = data;
	int i;

	fprintf(fp, "%s: manips %u\n", type, nat->num_manips);

	for (i = 0; i < nat->num_manips; i++) {
		struct ip_nat_info_manip *m = &nat->manips[i];
		struct ip_conntrack_manip *mip = &m->manip;
		
		fprintf(fp, "%s: ", hooks[m->hooknum]);
		fprintf(fp, "%s to ",
		        m->maniptype == IP_NAT_MANIP_SRC ? "src" : "dest");
		fprintf(fp, "%s:", inet_ntoa((struct in_addr){mip->ip}));
		/* FIXME: don't know proto */
		fprintf(fp, "%u", ntohs(mip->u.tcp.port));
		fprintf(fp, "\n");
	}

	return 0;
}

static int print_timeout(FILE *fp, char *type, void *data)
{
	fprintf(fp, "%s: +%lus\n", type, 100 * *(unsigned long *)data);
	return 0;
}

static int print_mark(FILE *fp, char *type, void *data)
{
	fprintf(fp, "%s: %lu\n", type, *(unsigned long *)data);
	return 0;
}


static int ctnl_parse_attr(struct nfattr *tb[], int max, struct nfattr *cta,
                           int len)
{
	memset(tb, 0, sizeof(struct nfattr *) * max);
	while (NFA_OK(cta, len)) {
		if (cta->nfa_type <= max)
			tb[cta->nfa_type] = cta;
		cta = NFA_NEXT(cta,len);
	}
	return 0;
}

int print_nlmsg(FILE *fp, struct nlmsghdr *n)
{
	struct nfgenmsg *cm = NLMSG_DATA(n);
	struct nfattr *cb[CTA_MAX+1];
	unsigned int type = NFNL_MSG_TYPE(n->nlmsg_type);
	int len = n->nlmsg_len;
	int i;

	len -= NLMSG_LENGTH(sizeof(*cm));
	if (len < 0) {
		fprintf(stderr, "BUG: wrong nlmsg len %d\n", len);
		return -1;
	}
	
	ctnl_parse_attr(cb, CTA_MAX, NFM_NFA(cm), len);

	switch (NFNL_MSG_TYPE(n->nlmsg_type)) {
	case CTNL_MSG_NEWCONNTRACK:
		if (n->nlmsg_flags & NLM_F_EXCL)
			fprintf(fp, "NEW:\n");
		else
			fprintf(fp, "CHANGE:\n");
		break;
	case CTNL_MSG_DELCONNTRACK:
		fprintf(fp, "DELETE:\n");
		break;
	default:
		fprintf(fp, "unknown message type 0x%x\n", type);
		return -1;
	}

	for (i = 0; i < CTA_MAX; i++) {
		if (cb[i] == NULL)
			continue;
		if (cta_table[i].name == NULL) {
			fprintf(fp,
			        "unknown attribut received: %u (%u bytes)\n",
			        i, NFA_PAYLOAD(cb[i]));
			continue;
		}
		if (NFA_PAYLOAD(cb[i]) < cta_table[i].size) {
			fprintf(fp, "short attribut: %s (%u/%u bytes)\n",
			        cta_table[i].name, NFA_PAYLOAD(cb[i]),
				cta_table[i].size);
			continue;
		}
		cta_table[i].print(fp, cta_table[i].name, NFA_DATA(cb[i]));
	}
	fprintf(fp, "\n");

	return 0;
}
