/*
 * utils.c      "ct" common helper functions
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Authors:     Patrick McHardy, <kaber@trash.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "ct.h"
#include "utils.h"

int get_integer(int *val, char *arg, int base)
{
	long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;
	res = strtol(arg, &ptr, base);
	if (!ptr || ptr == arg || *ptr || res > INT_MAX || res < INT_MIN)
		return -1;
	*val = res;
	return 0;
}

int get_unsigned(unsigned *val, char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;
	res = strtoul(arg, &ptr, base);
	if (!ptr || ptr == arg || *ptr || res > UINT_MAX)
		return -1;
	*val = res;
	return 0;
}

int get_ulong(unsigned long *val, char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;
	res = strtoul(arg, &ptr, base);
	if (!ptr || ptr == arg || *ptr || res > ULONG_MAX)
		return -1;
	*val = res;
	return 0;
}

int get_u32(__u32 *val, char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;
	res = strtoul(arg, &ptr, base);
	if (!ptr || ptr == arg || *ptr || res > 0xFFFFFFFFUL)
		return -1;
	*val = res;
	return 0;
}

int get_u16(__u16 *val, char *arg, int base)
{
	unsigned long res;
	char *ptr;
	
	if (!arg || !*arg)
		return -1;
	res = strtoul(arg, &ptr, base);
	if (!ptr || ptr == arg || *ptr || res > 0xFFFF)
		return -1;
	*val = res;
	return 0;
}

int get_addr(__u32 *addr, char *name)
{
	struct hostent *host = gethostbyname(name);

	if (host && host->h_addrtype == AF_INET) {
		*addr = *(__u32*)host->h_addr_list[0];
		return 0;
	}
	return -1;
}

int get_proto(__u16 *val, char *name)
{
	struct protoent *proto = getprotobyname(name);

	if (proto) {
		*val = proto->p_proto;
		return 0;
	}
	return -1;
}

int get_size(unsigned int *size, char *str)
{
	double sz;
	char *p;

	sz = strtod(str, &p);
	if (p == str)
		return -1;

	if (*p) {
		if (strcasecmp(p, "kb") == 0 || strcasecmp(p, "k")==0)
			sz *= 1024;
		else if (strcasecmp(p, "mb") == 0 || strcasecmp(p, "m")==0)
			sz *= 1024*1024;
		else if (strcasecmp(p, "mbit") == 0)
			sz *= 1024*1024/8;
		else if (strcasecmp(p, "kbit") == 0)
			sz *= 1024/8;
		else if (strcasecmp(p, "b") != 0)
			return unknown_arg(p);
	}

	*size = sz;
	return 0;
}

static int get_manip(struct ip_conntrack_manip *manip, int argc, char **argv)
{
	int eaten = 0;
	int ip = 0, port = 0;

	while (argc > 0) {
		if (matches(*argv, "host")) {
			_NEXT_ARG();
			if (get_addr(&manip->ip, *argv))
				return -1;
			ip = 1;
		} else if (matches(*argv, "port")) {
			_NEXT_ARG();
			if (get_u16(&manip->u.tcp.port, *argv, 0))
				return -1;
			port = 1;
/*
		} else if (matches(*argv, "icmp-type")) {
			_NEXT_ARG();
			if (get_u8(&manip->u.icmp.type, *argv, 0))
				return -1;
			port = 1;
		} else if (matches(*argv, "icmp-code")) {
			_NEXT_ARG();
			if (get_u8(&manip->u.icmp.code, *argv, 0))
				return -1;
*/
		} else
			break;
		argc--, argv++, eaten++;
	}

	if (!(ip & port))
		eaten = -1;

	return eaten;
}

int get_tuple(struct ip_conntrack_tuple *tuple, int argc, char **argv)
{
	int eaten = 0;
	int proto = 0, src = 0, dst = 0;

	while (argc > 0) {
		int i = 1;
		if (matches(*argv, "proto")) {
			_NEXT_ARG();
			if (get_proto(&tuple->dst.protonum, *argv))
				return -1;
			proto = 1;
		} else if (matches(*argv, "src")) {
			_NEXT_ARG();
			if ((i = get_manip(&tuple->src, argc, argv)) < 0)
				return -1;
			src = 1;
		} else if (matches(*argv, "dst")) {
			_NEXT_ARG();
			if ((i = get_manip((struct ip_conntrack_manip *)
			                   &tuple->dst, argc, argv)) < 0)
				return -1;
			dst = 1;
		} else
			break;

		argc -= i, argv += i, eaten += i;
	}

	if (!(src & dst & proto))
		eaten = -1;

	return eaten;
}

int get_status(unsigned long *val, char *arg)
{
	char *s = arg, *token;
	int ret = -1;

	*val = 0;
	while ((token = strsep(&s, ","))) {
		if (matches(token, "expected"))
			*val |= IPS_EXPECTED;
		else if (matches(token, "confirmed"))
			*val |= IPS_CONFIRMED;
		else if (matches(token, "replied"))
			*val |= IPS_SEEN_REPLY;
		else if (matches(token, "assured"))
			*val |= IPS_ASSURED;
		else
			return unknown_arg(token);
		ret = 0;
	}
	return ret;
}

static int get_state(unsigned int *state, char *name)
{
	static char *tcp_states[TCP_CONNTRACK_MAX + 1] = {
		"NONE",
		"ESTABLISHED",
		"SYN_SENT",
		"SYN_RECV",
		"FIN_WAIT",
		"TIME_WAIT",
		"CLOSE",
		"CLOSE_WAIT",
		"LAST_ACK",
		"LISTEN",
		"INVALID"
	};
	int i;

	for (i = 0; i <= TCP_CONNTRACK_MAX; i++)
		if (matches(name, tcp_states[i])) {
			*state = i;
			return 0;
		}

	return unknown_arg(name);
}
			

int get_protoinfo(struct cta_proto *cp, int argc, char **argv)
{
	int eaten = 0;
	int proto = 0;

	while (argc > 0) {
		if (matches(*argv, "proto")) {
			_NEXT_ARG();
			if (get_proto((__u16*)&cp->num_proto, *argv)) {
				printf("get_proto failed\n");
				return -1;
			}
			proto = cp->num_proto;
		} else if (matches(*argv, "handshake-ack")) {
			u_int32_t ack;
			if (proto != IPPROTO_TCP)
				return -1;
			_NEXT_ARG();
			if (get_u32(&ack, *argv, 0))
				return -1;
			cp->proto.tcp.handshake_ack = htonl(ack);
		} else if (matches(*argv, "state")) {
			if (proto != IPPROTO_TCP)
				return -1;
			_NEXT_ARG();
			if (get_state(&cp->proto.tcp.state, *argv))
				return -1;
		} else if (matches(*argv, "count")) {
			if (proto != IPPROTO_ICMP)
				return -1;
			_NEXT_ARG();
			if (get_unsigned((__u32*)&cp->proto.icmp.count, *argv, 0))
				return -1;
		} else
			break;

		argc--, argv++, eaten++;
	}

	return eaten;
}

int get_helpinfo(struct cta_help *ch, int argc, char **argv)
{
	return 0;
}

int get_natinfo(struct cta_nat *cn, int argc, char **argv)
{
	return 0;
}

int get_nlmcgroups(__u32 *val, char *arg)
{
	char *s = arg, *token;
	int ret = -1;

	*val = 0;
	while ((token = strsep(&s, ","))) {
		if (matches(token, "tcp"))
			*val |= NFGRP_IPV4_CT_TCP;
		else if (matches(token, "udp"))
			*val |= NFGRP_IPV4_CT_UDP;
		else if (matches(token, "icmp"))
			*val |= NFGRP_IPV4_CT_ICMP;
		else if (matches(token, "other"))
			*val |= NFGRP_IPV4_CT_OTHER;
		else
			return unknown_arg(token);
		ret = 0;
	}
	return ret;
}

unsigned int get_sock_rmem_max(void)
{
	int fd;
	char buf[64], *end;
	
	if ((fd = open("/proc/sys/net/core/rmem_max", O_RDONLY)) < 0)
		return 0;
	if (read(fd, buf, sizeof(buf)) < 0)
		return 0;
	close(fd);
	return strtol(buf, &end, 0);
}
	
int ctnl_parse_attr(struct nfattr *tb[], int max, struct nfattr *cta, int len)
{
	memset(tb, 0, sizeof(struct nfattr *) * max);
	while (NFA_OK(cta, len)) {
		if (cta->nfa_type <= max)
			tb[cta->nfa_type] = cta;
		cta = NFA_NEXT(cta,len);
	}
	return 0;
}

struct nlmsghdr *nfnl_next_msg(struct nlmsghdr **last_msg, char *buf,
                               size_t len)
{
	struct nlmsghdr *nlh;
	size_t remain_len;

	if ((char *)(*last_msg) > buf + len ||
	    (char *)(*last_msg) < buf)
		*last_msg = NULL;

	if (!*last_msg) {
		nlh = (struct nlmsghdr *)buf;
		if (!NLMSG_OK(nlh, len)) {
			fprintf(stderr, "error parsing nlmsg\n");
			return NULL;
		}
	} else {
		/* we are n-th part of multipart mesasge */
		if ((*last_msg)->nlmsg_type == NLMSG_DONE ||
		    !((*last_msg)->nlmsg_flags & NLM_F_MULTI)) {
			*last_msg = NULL;
			return NULL;
		}

		remain_len = len - ((char *)(*last_msg) - buf);
		nlh = NLMSG_NEXT(*last_msg, remain_len);
	}

	*last_msg = NULL;
	return nlh;
}

int matches(char *cmd, char *pattern)
{
	int len = strlen(cmd);
	if (len > strlen(pattern))
		return 0;
	return !strncasecmp(pattern, cmd, len);
}
